# coding: utf-8

"""
    Phantasma API

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: v1
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git
"""

import pprint
import re  # noqa: F401

import six

class StorageResult(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'available': 'int',
        'used': 'int',
        'avatar': 'str',
        'archives': 'list[ArchiveResult]'
    }

    attribute_map = {
        'available': 'available',
        'used': 'used',
        'avatar': 'avatar',
        'archives': 'archives'
    }

    def __init__(self, available=None, used=None, avatar=None, archives=None):  # noqa: E501
        """StorageResult - a model defined in Swagger"""  # noqa: E501
        self._available = None
        self._used = None
        self._avatar = None
        self._archives = None
        self.discriminator = None
        if available is not None:
            self.available = available
        if used is not None:
            self.used = used
        if avatar is not None:
            self.avatar = avatar
        if archives is not None:
            self.archives = archives

    @property
    def available(self):
        """Gets the available of this StorageResult.  # noqa: E501


        :return: The available of this StorageResult.  # noqa: E501
        :rtype: int
        """
        return self._available

    @available.setter
    def available(self, available):
        """Sets the available of this StorageResult.


        :param available: The available of this StorageResult.  # noqa: E501
        :type: int
        """

        self._available = available

    @property
    def used(self):
        """Gets the used of this StorageResult.  # noqa: E501


        :return: The used of this StorageResult.  # noqa: E501
        :rtype: int
        """
        return self._used

    @used.setter
    def used(self, used):
        """Sets the used of this StorageResult.


        :param used: The used of this StorageResult.  # noqa: E501
        :type: int
        """

        self._used = used

    @property
    def avatar(self):
        """Gets the avatar of this StorageResult.  # noqa: E501


        :return: The avatar of this StorageResult.  # noqa: E501
        :rtype: str
        """
        return self._avatar

    @avatar.setter
    def avatar(self, avatar):
        """Sets the avatar of this StorageResult.


        :param avatar: The avatar of this StorageResult.  # noqa: E501
        :type: str
        """

        self._avatar = avatar

    @property
    def archives(self):
        """Gets the archives of this StorageResult.  # noqa: E501


        :return: The archives of this StorageResult.  # noqa: E501
        :rtype: list[ArchiveResult]
        """
        return self._archives

    @archives.setter
    def archives(self, archives):
        """Sets the archives of this StorageResult.


        :param archives: The archives of this StorageResult.  # noqa: E501
        :type: list[ArchiveResult]
        """

        self._archives = archives

    def to_dict(self):
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(StorageResult, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self):
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self):
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other):
        """Returns true if both objects are equal"""
        if not isinstance(other, StorageResult):
            return False

        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        """Returns true if both objects are not equal"""
        return not self == other
